package org.nst.pasientlink.authenticator;

/**
 * <p>Title: Authenticator </p>
 * <p>Description: The authenticator implements a two-factor authentification. In the first phase,
 * the authenticator receives application generated requests for authentication of a
 * specified user (ID). When the request is received, a single-use password
 * is sent using GSM Short Messaging Service to a GSM cell phone registered on the
 * specified user. In the second phase of the authentication process, a second request
 * is made by the application that generated the request for authentication, specifying the
 * user id and the single-use password /hash) collected from the user.
 * If both the single-use and user specified password hashes match the generated
 * single-use password and stored user password respectively, the authenticator service
 * will return a boolean true value indicating successful authentication. Otherwise a boolean false value
 * will be returned, indicating failed authentication. In addition, the second phase must be
 * executed within a specific timeframe after the first phase of the authentication process
 * was initiated, in order to achieve successful authentication.
 *
 * The single-use passwords are generated using a secure random generator, and are
 * held in-memory only, for a limited period of time. As soon as a authentication
 * request has been completed, the single-use password is disposed of. </p>
 * <p>Copyright: Copyright (c) 2002</p>
 * <p>Company: Norwegian Centre for Telemedicine </p>
 * @author Thomas Strandens
 * @version 1.0
 */

import java.util.Hashtable;
import java.io.InputStream;
import java.io.IOException;
import java.util.Properties;
import org.apache.log4j.Logger;

public class Authenticator {
   // constants
   public static final int AUTHENTICATE_SESSION_OK = 0;
   public static final int AUTHENTICATE_TIMED_OUT = 1;
   public static final int AUTHENTICATE_WRONG_SINGLEUSEPASSWORD = 2;
   public static final int AUTHENTICATE_UNKNOWN_ERROR = -1;

   protected static final String SECURE_RANDOM_ALGORITHM = "SHA1PRNG"; // algorithm to be used for secure random generator
   protected static final String PASSWORD_HASH_ALGORITHM = "SHA1";
   protected static final String CRYPTOGRAPHY_PROVIDER = "SUN"; // provider to be used for secure random generator

   // instance variables
   private static int timeoutMinutes = 5; // default timeout in minutes
   private static Hashtable requests; // holds authentication requests, keyed by the user id
   private static AuthenticatorStorage myStore; // handles storage of user id, password hash, and mobile station subscriber number (MSISDN)
   private static Utility myUtility; // class containing utility methods for sending SMS etc.
   private Logger logger; // logger

   /**
    * Constructor. Initializes and configures system.
    */

   public Authenticator() {
      this.requests = new Hashtable();
      this.myUtility = new Utility();
      this.myStore = new AuthenticatorStorage(myUtility.getConfiguration());
      this.myUtility.setAuthenticatorStorage(this.myStore);
      this.logger = myUtility.getLogger();

      try {
         this.timeoutMinutes = Integer.parseInt(this.myUtility.getConfiguration().getProperty("sms_message_timeout_minutes"));
      }
      catch(Exception e) {
         this.logger.error("Unable to parse sms message timeout from configuration. Defaulting to " + this.timeoutMinutes + " minutes");
      }
   }

   /**
    * requestAuthentication performs the first phase of authentication (see class
    * documentation for further details of the two-phase authentification process).
    *
    * @param myUserID String containing the user id for the user to be authenticated
    * @param myPasswordHash byte[] containing the hashed password for the user
    * @return boolean indicating whether the first phase of the authentication process
    *   completed successfully (true) or not (false)
    *
    * @todo maybe rewrite return value into an int in order to encode more
    *   semantics into the return value?
    */

   public boolean requestAuthentication(String myUserID, byte[] myPasswordHash) {
      Request myRequest;
      boolean requestCompleted = false;
      // verify that user exists
      if(myUserID != null &&
         myPasswordHash != null &&
         this.myStore.userExists(myUserID) &&
            myUtility.verifyPasswordHash(myUserID, myPasswordHash)) {
         // first check whether we still have any pending requests for this user
         myRequest = (Request) requests.get(myUserID);
         if(myRequest != null) {
            // requests pending for this user
            logger.info("User '" + myUserID + "' re-requested authentication (previous request pending)");
            if(!myRequest.isTimedOut()) {
               requestCompleted = true; // assume user still has a valid single use password
            }
            else { // existing request has timed out - just fail after removing request
               logger.info("Previous request for authentication from UserID '" + myUserID + "' has timed out");
               this.requests.remove(myUserID);
               // user must now re-request authentication
            }
         }
         else {
            // no request existed - generate request and send single use password
            logger.info("UserID '" + myUserID + "' requested authentication");
            String mySingleUsePassword = myUtility.generateSingleUsePassword();
            myRequest = new Request(this.timeoutMinutes, myUserID, mySingleUsePassword);
            if(myUtility.sendSMSNotification(myRequest, null)) {
               this.requests.put(myUserID, myRequest);
               requestCompleted = true;
            }
            else {
               logger.fatal("Unable to spawn thread to send GSM-SMS notification for UserID '" + myUserID + "'");
            }
         }
      }
      else { // perform logging
         if(myUserID == null) {
            logger.warn("Received request authentication request with empty UserID parameter");
         }
         else {
            if(!this.myStore.userExists(myUserID)) {
               logger.warn("Received requestAuthentication request for non-existent UserID '" + myUserID + "'");
            }
            if(myPasswordHash == null) {
               logger.warn("Request authentication failure for UserID '" + myUserID + "': emtpy password hash");
            }
            else {
               if(this.myStore.userExists(myUserID)) {
                  if(!myUtility.verifyPasswordHash(myUserID, myPasswordHash)) {
                     logger.info("Request authentication failure for UserID '" + myUserID + "': password mismatch");
                  }
               }
            }
         }
      }
      return requestCompleted;
   }

   /**
    * authenticate performs actual authentication, after an authentication
    * request has been generated for a user (see class documentation for
    * a description of two-phased authentification process).
    *
    * @param myUserID String containing the userid of the user to authenticate
    * @param mySingleUsePasswordHash byte[] containing the single-use password
    *    hash, the hashed value of the single use password collected from the user for this auth. request
    * @return int value 0 if the user was authenticated, 1 if the request had timed out, 2 if the single use password was wrong
    * @todo logging for both successes and failures
    */

   public int authenticate(String myUserID, byte[] mySingleUsePasswordHash) {
      int retVal = AUTHENTICATE_UNKNOWN_ERROR; // default
      Request myRequest = null;
      // get authentification request - if this does not exist
      // then authentication cannot continue. perform additional
      // parameter verification as well
      if(myUserID != null) {
         myRequest = (Request) this.requests.get(myUserID);
      }
      if(myRequest != null &&
         mySingleUsePasswordHash != null) { // looks good this far..
         // get the single-use passwordhash generated and stored in the
         // authentication request
         byte[] requestSingleUsePasswordHash = myUtility.hashString(myRequest.getSingleUsePassword());

         // we are now ready to perform all tests required to authenticate
         // the user.
         if(!myRequest.isTimedOut() && // verify that request is still valid, i.e. has not timed out
            this.myUtility.byteArrayCompare(mySingleUsePasswordHash, requestSingleUsePasswordHash)) { // verify single-use passwords
               // all conditions hold. remove request from requests collection and
               // indicate to the garbage collector that we want the request object deleted by setting it to null
               // so that this request cannot be used again. then set return value
               this.requests.remove(myUserID);
               myRequest = null;
               retVal = AUTHENTICATE_SESSION_OK;
               // perform logging
               logger.info("Authenticated UserID: '" + myUserID + "'");

               // request garbage collection
               System.gc();
         }
         else { // user not authenticated - perform logging and other administrative tasks
            if(myRequest.isTimedOut()) { // remove from queue if the request was timed out
               this.requests.remove(myUserID);
               myRequest = null;
               retVal = AUTHENTICATE_TIMED_OUT;
               // log timeout failure
               logger.info("Authentication request for UserID '" + myUserID + "': request had timed out");
            }
            if(!this.myUtility.byteArrayCompare(mySingleUsePasswordHash, requestSingleUsePasswordHash)) {
               logger.info("Authentication request for UserID '" + myUserID + "': wrong single use password");
               retVal = AUTHENTICATE_WRONG_SINGLEUSEPASSWORD;
            }
         }
      }
      else { // something wrong with this call - log everything we can gather
         if(myUserID == null) {
            logger.warn("Received authenticate session request with empty userID parameter");
         }
         else {
            if(myRequest == null) {
               logger.warn("Received out of order authenticate session request for UserID '" + myUserID + "' (requestAuthentication has not been called first)");
            }
            if(mySingleUsePasswordHash == null) {
               logger.warn("Received authenticate session request for UserID '" + myUserID + "' with emtpy single use password");
            }
         }
      }
      return retVal;
   }

   /**
    * changePassword changes the user specified password. Both the old and new hashed password
    * must be supplied. the password is only changed into the new password if the old
    * password matches the stored password for the user. The application invoking this method
    * should ensure that the new password was typed in correctly before calling this method.
    *
    * @param myUserID String containing user identifier
    * @param myOldPasswordHash byte[] containing the old password, hashed
    * @param myNewPasswordHash byte[] containing the new password, hashed
    * @return boolean indicating whether the password was changed (true) nor not (false). failure to
    *   march the old password with the the main
    *   reason for a failure (false retval)
    */

   public boolean changePassword(String myUserID, byte[] myOldPasswordHash, byte[] myNewPasswordHash) {
      boolean retVal = false;
      if(myUserID != null &&
         myOldPasswordHash != null &&
         myNewPasswordHash != null) {
            if(this.myUtility.verifyPasswordHash(myUserID, myOldPasswordHash)) {
               if(this.myStore.changePassword(myUserID, myNewPasswordHash)) {
                  retVal = true;
                  logger.info("Changed password for UserID '" + myUserID + "'");
               }
               else {
                  logger.error("Unable to change password for UserID '" + myUserID + "'");
               }
            }
            else {
               logger.info("Unable to change password for UserID '" + myUserID + "': old password mismatch");
            }
      }
      else {
         String nullObjects = "";
         if(myUserID == null) {
            nullObjects = "UserID";
         }
         if(myOldPasswordHash == null) {
            nullObjects = nullObjects + " old_password_hash";
         }
         if(myNewPasswordHash == null) {
            nullObjects = nullObjects + " new_password_hash";
         }
         logger.warn("Change password was called with null objects: " + nullObjects);
      }
      return retVal;
   }

   /**
    * Main method. For creating
    *
    * @param args string array containing command line parameters
    *
    */

   public static void main(String[] args) {
      Authenticator myAuth = new Authenticator();
      if(args.length == 5) {
         if(myAuth.myUtility.createUser(args[1], args[2], args[3], args[4])) {
            System.out.println("Created user.");
         }
         else {
            System.out.println("Failed to create user");
         }
      }
      System.exit(0);
   }
}