package org.nst.pasientlink.authenticator;

/**
 * <p>Title: SendSMSThread </p>
 * <p>Description: This thread sends an SMS Message using the gsmlib gsmsendsms command.
 * The thread synchronizes on an input object/semaphore, suppled from the Utility.sendsms method.</p>
 * <p>Copyright: Copyright (c) 2002</p>
 * <p>Company: Norwegian Centre for Telemedicine</p>
 * @author Thomas Strandens
 * @version 1.0
 * @todo remove logging of single use password when SMS sending fails! only relevant during development
 */

import org.apache.log4j.Logger;
import java.io.IOException;

class SendSMSThread extends Thread {

   private static String SENDSMS_MESSAGE_HEADER = Utility.getConfiguration().getProperty("sms_message_header");
   private static String SENDSMS_MESSAGE_FOOTER = Utility.getConfiguration().getProperty("sms_message_footer");
   private final static String SENDSMS_COMMAND_SH = "/bin/sh";
   private final static String SENDSMS_COMMAND_SH_PARAMETER = "-c";
   private final static String SENDSMS_EXTERNAL_COMMAND = Utility.getConfiguration().getProperty("gsmsendsms_command");
   private Object semaphore;
   private String singleUsePassword;
   private String GSMSubscriberID;
   private String userID;
   private Logger logger;

   /**
    * Constructor. Sets up object.
    *
    * @param mySemaphore Object object used for synchronizing access to GSM device
    * @param myRequest Request containing single use password and userid
    * @param myGSMSubscriberID String containing GSM-SMS subscriber id
    */

   public SendSMSThread(Object mySemaphore,
                        Request myRequest,
                        String myGSMSubscriberID) {
      super("SendSMSThread");
      this.semaphore = mySemaphore;
      this.singleUsePassword = myRequest.getSingleUsePassword();
      this.GSMSubscriberID = myGSMSubscriberID;
      this.userID = myRequest.getUserID();
      this.logger = Utility.getLogger();
   }

   /**
    * run synchronizes on the input semaphore object, and sends an SMS
    * using the configured device.
    */

   public void run() {
      if(semaphore != null &&
         singleUsePassword != null &&
         GSMSubscriberID != null) {

         synchronized(semaphore) {
            String SMSMessage = this.SENDSMS_MESSAGE_HEADER +
               " " +
               singleUsePassword +
               this.SENDSMS_MESSAGE_FOOTER;
            // execute external command in order to send SMS Message
            try {
               Process myProcess = Runtime.getRuntime().exec(new String[] {this.SENDSMS_COMMAND_SH,
                                                                           this.SENDSMS_COMMAND_SH_PARAMETER,
                                                                           this.SENDSMS_EXTERNAL_COMMAND +
                                                                           " " +
                                                                           GSMSubscriberID +
                                                                           " '" +
                                                                           SMSMessage +
                                                                           "'"});
               int exitValue = myProcess.waitFor();
               if(exitValue == 0) { // 0 indicates success
                  logger.info("Completed sending of GSM-SMS for UserID '" + userID + "' to subscriber_id " + GSMSubscriberID);
               }
               else {
                  logger.fatal("Failed sending GSM-SMS to UserID '" + userID + "', subscriber_id " + GSMSubscriberID + ". Exit val from gsmsendsms-command: " + exitValue);
               }
            }
            catch(IOException e) {
               // error sending GSM SMS - critical
               logger.fatal("Failed sending GSM-SMS message: " + e.getMessage());
            }
            catch(InterruptedException e) { // only happens if someone calls 'interrupt' on this thread
               // potential error sending GSM SMS - critical
               logger.fatal("Failed sending GSM-SMS message: " + e.getMessage());
            }
         }
      }
      else { // not sufficient information to send sms - error situation
         if(semaphore == null) {
            logger.fatal("Semaphore for sending SMS is null");
         }
         if(singleUsePassword == null) {
            logger.error("Unable to send GSM-SMS message: Single use password is null");
         }
         if(GSMSubscriberID == null) {
            logger.error("Unable to send GSM-SMS message: GSM Subscriber ID is null");
         }
      }
   }
}